/*
走行管理モジュール
・左右のタイヤ回転距離の差から速度指令値を変更し、直進走行および回転走行を行う。
・指定距離または角度に近づいたら減速し、指定した距離で停止する。

2014/07/14	by S.Ushimaru
*/

#include <stdio.h>
#include <linux/errno.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/io.h>
#include <math.h>
#include "hardware.h"
#include "vel_ctrl.h"
#include "distance.h"
#include "run_ctrl.h"

static pthread_t task_id;
static float run_speed;
static int dist_flag=0;
static float run_dist;
static int ctrl_endflag=0;

static enum run_mode run_mode=STRAIGHT;

// Kp_s : 直進制御において、左右のタイヤの距離差に対する補正速度の係数
static float Kp_s = 14.2;//20.0; 
static float Kd_s = 2.8;//25.5;//31.5; 
// Kp_r : 回転制御において、左右のタイヤの距離差に対する補正速度の係数
static float Kp_r = 7.0;//5.3; 
static float Kd_r = 3;//0.0;10.0; 

// Kf_s : 直進停止制御において、残り距離に対する速度指定値を決める係数
// Kf_r : 回転停止制御において、残り角度に対する速度指定値を決める係数
static float Kf_s = 5.0 ;//2.0; // 1.0
static float Kf_r = 0.25;
static float Kf_r_org ;
static float diff_l_r_prev = 0;
static float sum_diff_l_r = 0;

static void *thread_code(void *t){

	int counter=0;
	struct distance_data distance_data;

	float dist_vel_down = 5; //直進停止制御において、減速を開始する残り距離
	float angle_vel_down = 20; //回転停止制御において、減速を開始する残り角度

	float diff_l_r;
	float mod_vel;
	float curr_dist;
        float coef;
        float dist_diff;
	float vel_speed=0;

	//タイヤ走行距離を車体回転角度に変換する際の係数
	coef = 1. /BETWEEN_TIRE  * 180 / 3.14159;

	Kf_r_org = Kf_r;

	while(1){
		counter++;

		distance_get(&distance_data);
		//printf("dis_left=%6.1f    dis_r=%6.1f\n",distance_data.left, distance_data.right);
		//printf("ctrl_endflag=%d\n",ctrl_endflag);

		vel_speed = run_speed;	

		switch(run_mode){

		case STRAIGHT:
			//printf("case STRAIGHT    dist_flag=%d\n",dist_flag);
			if(dist_flag==1){
				curr_dist = ( distance_data.left + distance_data.right ) / 2.0;
				dist_diff = run_dist - curr_dist;
				if(run_speed >0){
					//指定距離に近づいたときに距離に応じて指定速度を小さくする
					if( dist_diff < dist_vel_down){
						//vel_speed = Kf_s * dist_diff + 2;
						//vel_speed = (run_speed - 2) * dist_diff / dist_vel_down + 2;
					  vel_speed = (run_speed -45);
          }
					//printf("curr_dist=%6.1f    run_dist=%6.1f\n",curr_dist, run_dist);
					if(curr_dist > run_dist) goto END;
				}else{
					//指定距離に近づいたときに距離に応じて指定速度を小さくする
					if( dist_diff > -dist_vel_down){
						//vel_speed = Kf_s * dist_diff - 2;
						vel_speed = -(run_speed + 2) * dist_diff / dist_vel_down - 2;
					}
					if(curr_dist < run_dist) goto END;
				}
				vel_ctrl_set(vel_speed, vel_speed);
			}

			//距離データの差に比例して左右の速度指令値を変更する
			diff_l_r = distance_data.left - distance_data.right;
			sum_diff_l_r += fabs(diff_l_r);
			//mod_vel = Kp_s * diff_l_r ;
			mod_vel = Kp_s * diff_l_r + Kd_s * (diff_l_r - diff_l_r_prev);
			diff_l_r_prev = diff_l_r;

			vel_ctrl_mod( -mod_vel, mod_vel);

			break;

		case ROTATE:
			//printf("case ROTATE    dist_flag=%d\n",dist_flag);
			if(dist_flag==1){
				curr_dist = ( distance_data.left - distance_data.right ) *coef;
				dist_diff = run_dist - curr_dist;
				if(run_speed >0){
					//指定角度に近づいたときに角度に応じて指定速度を小さくする
					if( dist_diff < angle_vel_down ){
						vel_speed = Kf_r * dist_diff + 2;
						//printf("vel speed = %6.1f  in spd>0\n",vel_speed);
					}
					//printf("curr_dist=%6.1f    run_dist=%6.1f\n",curr_dist, run_dist);
					if(curr_dist > run_dist) goto END;
				}else{
					//指定角度に近づいたときに角度に応じて指定速度を小さくする
					if( dist_diff > -angle_vel_down){
						vel_speed = Kf_r * dist_diff - 2;
						//printf("vel speed = %6.1f  in spd<0f\n",vel_speed);
					}
					//printf("curr_dist=%6.1f    run_dist=%6.1f\n",curr_dist, run_dist);
					if(curr_dist < run_dist) goto END;
				}
				vel_ctrl_set(vel_speed, -vel_speed);
			}

			//距離データの差に比例して左右の速度指令値を変更する
			diff_l_r = distance_data.left + distance_data.right;
			sum_diff_l_r += fabs(diff_l_r);
			//mod_vel = Kp_r * diff_l_r ;
			mod_vel = Kp_r * diff_l_r + Kd_r * (diff_l_r - diff_l_r_prev);
			//printf("diff_l_r=%5.1f   mod_vel=%5.1f\n",diff_l_r, mod_vel);
			//vel_ctrl_mod( -mod_vel, -mod_vel);
			break;

		default:
			 break;

		};


		usleep((T_RUN_CTRL - 1)*1000); // 走行制御（直進、回転、旋回）

	}

	END:
	printf("sum_diff_l_r=%5.1f\n",sum_diff_l_r);
	vel_ctrl_set(0., 0.);
	ctrl_endflag=1;

	return 0;

}


static int thread_create()
{
	pthread_attr_t attr;
	struct sched_param sched_param;
	int ret;
	int policy;

	pthread_attr_init(&attr);

	// 優先度20、ポリシーFIFOでスレッド生成
	sched_param.sched_priority = 20;
	pthread_attr_setschedparam(&attr, &sched_param);
	pthread_setschedparam(pthread_self(), SCHED_FIFO , &sched_param);

	ret = pthread_create(&task_id, &attr, thread_code, (void *)2);

	pthread_getschedparam(pthread_self(), &policy, &sched_param);
	printf("plicy = %s, priority =%d\n",
		(policy == SCHED_FIFO)  ? "SCHED_FIFO" :
		(policy == SCHED_RR)    ? "SCHED_RR" :
		(policy == SCHED_OTHER) ? "SCHED_OTHER" :
		"???",
		sched_param.sched_priority);

	return 0;
}


void run_ctrl_create(){

	ctrl_endflag = 0;
	diff_l_r_prev = 0;
	sum_diff_l_r = 0;
	distance_reset();
	thread_create();

}

int run_ctrl_get_endflag(){

	return ctrl_endflag;
}

void run_ctrl_cancel(){

	ctrl_endflag=0;
	pthread_cancel(task_id);

}

void run_ctrl_reset(){

	distance_reset();
	//ctrl_endflag=0;
	//dist_flag = 0;

}

void run_ctrl_set_dist(float dist){

	ctrl_endflag=0;
	dist_flag = 1;
	run_dist = dist;

}

void run_ctrl_unset_dist(){
	run_dist = 0;
	dist_flag = 0;
}

void run_ctrl_stop(){
	vel_ctrl_set(0., 0.);
}

void run_ctrl_set_Kfr(float Kfr){
	Kf_r  =  Kfr;
}

void run_ctrl_reset_Kfr(){
	Kf_r  =  Kf_r_org;
}

void run_ctrl_set(int mode, float speed){

	float rot_coef ;
	float vel_l_ref=0, vel_r_ref=0;

	rot_coef = BETWEEN_TIRE/2.0 * 3.1415 / 180.;
	run_mode = mode;

	run_speed = speed;

	switch(run_mode){
	case STRAIGHT:
		printf("straight mode selected\n");
		vel_l_ref=speed;
		vel_r_ref=speed;
		break;
	case ROTATE:
		printf("rotae mode selected\n");
		vel_l_ref =  speed * rot_coef;
		vel_r_ref = -speed * rot_coef;
		break;
	default:
		break;
	}

	vel_ctrl_set(vel_l_ref, vel_r_ref);

}

